from llm_strategy.llm_function import (  # type: ignore
    llm_explicit_function,
    llm_function,
)
from llm_strategy.llm_strategy import llm_dataclass, llm_strategy  # type: ignore

__version__ = "2.1.1"
